#![feature(concat_idents)]

use rand::random;

use hercules_rt::{runner, HerculesCPURef};
#[cfg(feature = "cuda")]
use hercules_rt::CUDABox;

juno_build::juno!("matmul");

fn main() {
    async_std::task::block_on(async {
        const I: usize = 256;
        const J: usize = 64;
        const K: usize = 128;
        let mut a: Box<[i32]> = (0..I * J).map(|_| random::<i32>() % 100).collect();
        let mut b: Box<[i32]> = (0..J * K).map(|_| random::<i32>() % 100).collect();
        let mut correct_c: Box<[i32]> = (0..I * K).map(|_| 0).collect();
        for i in 0..I {
            for k in 0..K {
                for j in 0..J {
                    correct_c[i * K + k] += a[i * J + j] * b[j * K + k];
                }
            }
        }
        #[cfg(not(feature = "cuda"))]
        {
            let a = HerculesCPURef::from_slice(&a);
            let b = HerculesCPURef::from_slice(&b);
            let mut r = runner!(matmul);
            let c = r.run(I as u64, J as u64, K as u64, a.clone(), b.clone()).await;
            assert_eq!(c.as_slice::<i32>(), &*correct_c);
            let mut r = runner!(tiled_64_matmul);
            let tiled_c = r.run(I as u64, J as u64, K as u64, a.clone(), b.clone()).await;
            assert_eq!(tiled_c.as_slice::<i32>(), &*correct_c);
        }
        #[cfg(feature = "cuda")]
        {
            let a = CUDABox::from_cpu_ref(HerculesCPURef::from_slice(&mut a));
            let b = CUDABox::from_cpu_ref(HerculesCPURef::from_slice(&mut b));
            let mut r = runner!(matmul);
            let c = r.run(I as u64, J as u64, K as u64, a.get_ref(), b.get_ref()).await;
            let mut c_cpu: Box<[i32]> = vec![0; correct_c.len()].into_boxed_slice();
            c.to_cpu_ref(&mut c_cpu);
            assert_eq!(&*c_cpu, &*correct_c);
            let mut r = runner!(tiled_64_matmul);
            let tiled_c = r.run(I as u64, J as u64, K as u64, a.get_ref(), b.get_ref()).await;
            let mut tiled_c_cpu: Box<[i32]> = vec![0; correct_c.len()].into_boxed_slice();
            tiled_c.to_cpu_ref(&mut tiled_c_cpu);
            assert_eq!(&*tiled_c_cpu, &*correct_c);
        }
    });
}

#[test]
fn matmul_test() {
    main();
}
